/* Set up ALSA configuration
 * Martin Stransky <stransky@redhat.com>
 *
 * Copyright (C) 2003,2005 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <pwd.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <time.h>

#include "ainit.h"
#include "ini.h"
#include "shm.h"

static void usage(FILE * out)
{
  fprintf(out, "Usage: ainit user_name start|stop [--version] [--help]\n\n");
}

static void version(void)
{
  printf("ALSA Configuration tool, Copyright (C) 2003,2005 Red Hat, Inc.\n"
         "This is free software; see the source for copying conditions.\n"
         "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n");
  printf("Version %s\n\n", VERSION);
}

static void help(void)
{
  version();
  usage(stdout);
}

static void bad_param(char *p_name, char *p_par)
{
  usage(stderr);
  fprintf(stderr, "Bad parameter - %s", p_par);
  exit(1);
}

/* Print error message and/or end
*/
void error(int end, int err, char *p_text, ...)
{
#define   TEXTLEN 2000
  char text[TEXTLEN];
  va_list args;

  int errno_back = errno;

  if (p_text) {
    va_start(args, p_text);
    vsnprintf(text, TEXTLEN, p_text, args);
    va_end(args);
    syslog(3, text);
  }

  if (err)
    syslog(3, "Error: %s", strerror(errno_back));

  if (end)
    exit(1);
}

void file_copy(char *p_src, char *p_dest)
{

#define COPY_BUFFER 4096

  FILE *src = NULL, *dest = NULL;
  char  buff[COPY_BUFFER];
  size_t ret;

  if (!(src = fopen(p_src,"rb")))
      goto err;
  if (!(dest = fopen(p_dest,"wb")))
      goto err;
  
  while((ret = fread(buff,sizeof(char), sizeof(buff), src)))      
      fwrite(buff,sizeof(char),ret,dest);
  
err:
  if(errno) {
    syslog(3,strerror_r(errno, buff, sizeof(buff)));
  }
  if(src)
      fclose(src);
  if(dest)
      fclose(dest);
}

/* Replace '~' and '$HOME' with *p_home string
*/
static void parse_path_to_file(char *p_in, char *p_out, size_t max_len,
                               char *p_home)
{
  const char *p_prefix;

  if (is_token(p_in, TOKEN_HOME1)) {
    p_prefix = p_home;
    p_in += strlen(TOKEN_HOME1);
  }
  else if (is_token(p_in, TOKEN_HOME2)) {
    p_prefix = p_home;
    p_in += strlen(TOKEN_HOME2);
  }
  else {
    p_prefix = "";
  }

  if (strlen(p_prefix) + strlen(p_in) < max_len) {
    strcpy(p_out, p_prefix);
    strcat(p_out, p_in);
  }
  else {
    error(TRUE, FALSE, "Too long filename %s%s.", p_prefix, p_in);
  }
}

/* Setting rights for root/user
*/
static void set_rights_user(CNF * p_cnf)
{
  setegid(p_cnf->p_pass->pw_gid);
  seteuid(p_cnf->p_pass->pw_uid);
}

static void set_rights_root(CNF * p_cnf)
{
  setegid(ROOT_GID);
  seteuid(ROOT_UID);
}

static void copy_comment(FILE * f_in, FILE * f_out)
{
  int c;

  while ((c = fgetc(f_in)) != EOF) {
    fputc(c, f_out);
    if (c == '\n')
      break;
  }
}

/* Parse template file (p_in), allocate mem/semaphores
   and write results to target file (f_out)
*/
static void parse_file(CNF * p_cnf, FILE * f_in, FILE * f_out)
{
  char line[MAX_TOKEN_LEN];
  int c;

  while ((c = fgetc(f_in)) != EOF) {
    if (c == SEPARATOR) {
      if (read_token(f_in, line, MAX_TOKEN_LEN, SEPARATOR)) {
        if (is_token(line, IPC_MEM_DIRECT)) {
          int key = alloc_ipc_key();
          fprintf(f_out, "%s %d", IPC_MEM_DIRECT, key);
        }
        else if (is_token(line, IPC_MEM_PLUG)) {
          int key = alloc_ipc_key();
          fprintf(f_out, "%s %d", IPC_MEM_PLUG, key);
        }
        else if (is_token(line, IPC_SEM)) {
          int key = alloc_ipc_semaphore();
          fprintf(f_out, "%s %d", IPC_SEM, key);
        }
        else {
          error(FALSE, FALSE, "Uknown token %s", line);
        }
      }
      else {
        fputc(c, f_out);
        fputs(line, f_out);
      }
    }
    else {
      fputc(c, f_out);
      if (c == COMMENT)
        copy_comment(f_in, f_out);
    }
  }
}

/* Parse teplate file, it's determined by num parameter
*/
static int process_template(CNF * p_cnf, FILE * file_config, int num)
{
  char template[FILE_LEN];
  char target[FILE_LEN];
  char rights[FILE_LEN];
  char tmp[FILE_LEN];
  FILE *f_in, *f_out;

  /* Making tokens from config file
   */
  snprintf(template, FILE_LEN, "%s_%d", OPT_TEMPLATE, num);
  snprintf(target, FILE_LEN, "%s_%d", OPT_TARGET, num);
  snprintf(rights, FILE_LEN, "%s_%d", OPT_TARGET_ROOT, num);

  /* Reading tokens from config file
     and parse it
   */
  ini_read_string(file_config, template, tmp, FILE_LEN, "");
  if (tmp[0])
    parse_path_to_file(tmp, template, FILE_LEN, p_cnf->p_pass->pw_dir);
  else
    return (FALSE);

  ini_read_string(file_config, target, tmp, FILE_LEN, "");
  if (tmp[0])
    parse_path_to_file(tmp, target, FILE_LEN, p_cnf->p_pass->pw_dir);
  else
    return (FALSE);

  /* Making name of lock file
   */
  if (strlen(target) + strlen(FILE_LOCK) < FILE_LEN) {
    strcpy(tmp, target);
    strcat(tmp, FILE_LOCK);
  }
  else {
    error(TRUE, FALSE, "Too long file name %s", target);
  }

  /* Set root right for a target file (if it's requied)
   */
  if (ini_read_bool(file_config, rights, 0))
    set_rights_root(p_cnf);

  /* Opening template and target file
   */
  if ((f_in = fopen(template, "r")) == NULL) {
    error(TRUE, TRUE, "Failed to open template file %s", template);
  }
  if (!(p_cnf->overwrite) && (f_out = fopen(target, "r")) != NULL) {
    error(TRUE, TRUE, "File %s already exists.", target);
  }
  else if ((f_out = fopen(target, "w")) == NULL) {
    error(TRUE, TRUE, "Failed to open file %s", target);
  }

  set_rights_user(p_cnf);


  /* Processing those files
   */
  parse_file(p_cnf, f_in, f_out);


  /* Finishing
   */
  set_rights_root(p_cnf);

  fclose(f_in);
  fclose(f_out);

  /* Creating lock file
   */

  if ((f_out = fopen(tmp, "w")) == NULL) {
    error(TRUE, TRUE, "Failed to open file %s", tmp);
  }
  else {
    fclose(f_out);
  }

  set_rights_user(p_cnf);

  return (TRUE);
}

/* Create a config files (if ainit has the 'start' command line parameter)
*/
static void create_alsa_config(CNF * p_cnf)
{
  FILE *f;
  int i;

  if ((f = fopen(FILE_CONFIG, "r")) == NULL)
    error(TRUE, TRUE, "Failed to open file %s", FILE_CONFIG);

  for (i = 0; i < TEMPLATENUM; i++) {
    process_template(p_cnf, f, i);
  }

  fclose(f);
}

/* Parse target file and free semaphores/mem
*/
static void parse_file_back(FILE * f_in)
{
  char line[MAX_TOKEN_LEN];
  char *p_akt;
  char *p_tmp;
  int key;
  int IPC_SEM_LEN = strlen(IPC_SEM);
  int IPC_MEM_DIRECT_LEN = strlen(IPC_MEM_DIRECT);
  int IPC_MEM_PLUG_LEN = strlen(IPC_MEM_PLUG);

  while (fgets(line, MAX_TOKEN_LEN, f_in)) {
    p_akt = line;
    while (p_akt) {
      if ((p_tmp = strstr(p_akt, IPC_MEM_DIRECT))) {
        sscanf(p_tmp + IPC_MEM_DIRECT_LEN, "%d", &key);
        free_ipc_key(key);
        p_akt = p_tmp + IPC_MEM_DIRECT_LEN;
      }
      if ((p_tmp = strstr(p_akt, IPC_MEM_PLUG))) {
        sscanf(p_tmp + IPC_MEM_PLUG_LEN, "%d", &key);
        free_ipc_key(key);
        p_akt = p_tmp + IPC_MEM_PLUG_LEN;
      }
      else if ((p_tmp = strstr(p_akt, IPC_SEM))) {
        sscanf(p_tmp + IPC_SEM_LEN, "%d", &key);
        free_ipc_semaphore(key);
        p_akt = p_tmp + IPC_SEM_LEN;
      }
      else {
        p_akt = NULL;
      }
    }
  }
}

/* Parse a target file, free mem/semaphores and delete it
*/
static int process_template_back(CNF * p_cnf, FILE * file_config, int num, int start)
{
  char target_back[FILE_LEN];
  char target[FILE_LEN];
  char tmp[FILE_LEN];
  FILE *f_in;

  /* Making tokens from config file
   */
  snprintf(target, FILE_LEN, "%s_%d", OPT_TARGET, num);
  snprintf(target_back, FILE_LEN, "%s_%d", OPT_TARGET_BACK, num);

  /* Reading tokens from config file
     and parse it
   */
  ini_read_string(file_config, target, tmp, FILE_LEN, "");
  if (tmp[0])
    parse_path_to_file(tmp, target, FILE_LEN, p_cnf->p_pass->pw_dir);
  else
    return (FALSE);

  /* Making name of lock file
   */
  if (strlen(target) + strlen(FILE_LOCK) < FILE_LEN) {
    strcpy(tmp, target);
    strcat(tmp, FILE_LOCK);
  }
  else {
    error(TRUE, FALSE, "Too long file name %s", target);
  }

  set_rights_root(p_cnf);

  /* Checking for lock file
   */
  if ((f_in = fopen(tmp, "r")) == NULL) {
    /* Don't remove it and go to next template...
     */
    return (TRUE);
  }
  else {
    fclose(f_in);
    remove(tmp);
  }

  /* Removing target file
   */
  if ((f_in = fopen(target, "r")) != NULL) {
    if(!start)
      parse_file_back(f_in);
    fclose(f_in);
    remove(target);
  }
  else {
      error(FALSE, TRUE, "Failed to open file %s", p_cnf->file_target);
  }

  /* Restoring original config from back-up
  */
  ini_read_string(file_config, target_back, tmp, FILE_LEN, "");
  if (tmp[0]) {
    parse_path_to_file(tmp, target_back, FILE_LEN, p_cnf->p_pass->pw_dir);
    file_copy(target_back, target);
  }

  set_rights_user(p_cnf);

  return (TRUE);
}

/* Remove a config files (if ainit is runnig with the 'stop' 
   command line parameter)
*/
static void remove_alsa_config(CNF * p_cnf, int start)
{
  FILE *f;
  int i;

  set_rights_root(p_cnf);

  if ((f = fopen(FILE_CONFIG, "r")) == NULL) {
    error(FALSE, TRUE, "Failed to open file %s", FILE_CONFIG);
    return;
  }

  for (i = 0; i < TEMPLATENUM; i++) {
    process_template_back(p_cnf, f, i, start);
  }

  fclose(f);
}

static void load_and_set_config(CNF * p_cnf, char *p_user_name)
{
  FILE *f;

  memset(p_cnf, 0, sizeof(p_cnf[0]));

  if ((f = fopen(FILE_CONFIG, "r")) == NULL) {
    error(TRUE, TRUE, "Failed to open file %s", FILE_CONFIG);
  }
  else {
    p_cnf->enable = ini_read_bool(f, OPT_ENABLE, 0);
    p_cnf->overwrite = ini_read_bool(f, OPT_OVERWRITE, 0);

    p_cnf->p_pass = getpwnam(p_user_name);
    if (!p_cnf->p_pass)
      error(TRUE, TRUE, NULL);

    fclose(f);

    set_rights_user(p_cnf);
  }
}

/* Main loop - ainit user_name start|stop [--version] [--help]
*/
int main(int argc, char **argv)
{
  CNF cnf;

  if (argc < 3 || (argc > 1 && (is_token(argv[1], "--help") || is_token(argv[1], "-h") ||
      is_token(argv[1], "-?") || is_token(argv[1], "--version")))) {
    help();
    exit(0);
  }

  load_and_set_config(&cnf, argv[1]);

  if (is_token(argv[2], PARAM_START)) {
    remove_alsa_config(&cnf, TRUE);
    if(cnf.enable)
      create_alsa_config(&cnf);
  }
  else if (is_token(argv[2], PARAM_STOP)) {
    remove_alsa_config(&cnf, FALSE);
  }
  else {
    bad_param(argv[0], argv[2]);
  }

  return(0);
}
