/* Set up ALSA configuration
 * Martin Stransky <stransky@redhat.com>
 *
 * Copyright (C) 2003,2005 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* SHM functions
*/

#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

#include "shm.h"

void error(int end, int err, char *p_text, ...);

/* Checking if IPC key/semaphore already exist
*/
int check_ipc_key(key_t key)
{
  int shm_id = shmget(key, 1, IPC_CREAT | IPC_EXCL | 0600);
  return (shm_id >= 0);
}

int check_ipc_semaphore(key_t key)
{
  int shm_id = semget(key, 1, IPC_CREAT | IPC_EXCL | 0600);
  return (shm_id >= 0);
}

/* Random generator for IPC keys
*/
#define KEY_MIN  2000
#define KEY_MAX (INT_MAX)

key_t generate_key(int (*p_func) (key_t key))
{
  key_t key;
  int i;
  static int first = TRUE;

  if (first) {
    srand(time(NULL));
    first = FALSE;
  }

  for (i = rand(); i < INT_MAX; i++) {
    if (i < KEY_MIN)
      i += KEY_MIN;
    if (i > KEY_MAX) {
      i = rand();
      continue;
    }

    key = i;

    if (p_func(key))
      return (key);
    else {
      key = -i;
      if (p_func(key))
        return (key);
    }
  }
  return (FALSE);
}

/* Allocating shared memmory segment
*/
key_t alloc_ipc_key(void)
{
  key_t key;

  key = generate_key(check_ipc_key);
  if (!key) {
    error(FALSE, TRUE, "Memory: Failed to allocate SHM segment, set to 0");
  }

  return (key);
}

/* Freeing shared memmory segment
*/
void free_ipc_key(key_t key)
{
  int shm_id = shmget(key, 0, 0600);
  if (shm_id == -1 || shmctl(shm_id, IPC_RMID, NULL) < 0)
    error(FALSE, TRUE, "Memory: Failed to release SHM segment");
}

/* Allocating IPC semaphore
*/
key_t alloc_ipc_semaphore(void)
{
  key_t sem;

  sem = generate_key(check_ipc_semaphore);
  if (!sem) {
    error(FALSE, TRUE, "Semaphore: Failed to allocate semaphore, set to 0");
  }

  return (sem);
}

/* Freeing IPC semaphore
*/
void free_ipc_semaphore(key_t key)
{
  int shm_id = semget(key, 1, 0600);
  if (shm_id == -1 || semctl(shm_id, 0, IPC_RMID, NULL) < 0)
    error(FALSE, TRUE, "Memory: Failed to release semaphore");
}
